// Copyright 1996, 1997 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "constant.h"
#include "debug.h"

static void addXRay(const Board &board, Bitmap & attacks,Square attack_square,
                    Square square,ColorType side)
{
   int dir = Bearing::Directions[attack_square][square];
   if (dir == 0)
      return;
   const int r = ((int)attack_square)/8;
   const int f = File(attack_square)-1;
   switch (dir)
   {
   case 1:
      // Find attacks on the rank
      // shift the occupied bits so that the rank that 
      // attack_square occupies is in bits 0-7.
      {
         Bitmap occupied(board.occupied[White]);
         occupied.or(board.occupied[Black]);
         byte atcks = occupied.rank_bits(r);
         // find the next piece on the rank
         Bitmap occ((uint64)Bearing::next_bit_left[atcks][f]);
         // shift back up
         occ.shl_byte(r);
         // See if any rooks or queens are there
         occ.and(board.rooksqueens[side]);
         attacks.or(occ);
      }
      break;
   case -1:
      {
         // Find attacks on the rank
         // shift the occupied bits so that the rank that 
         // attack_square occupies is in bits 0-7.
         Bitmap occupied(board.occupied[White]);
         occupied.or(board.occupied[Black]);
         byte atcks = occupied.rank_bits(r);
         // find the next piece on the rank
         Bitmap occ((uint64)Bearing::next_bit_right[atcks][f]);
         // shift back up
         occ.shl_byte(r);
         // See if any rooks or queens are there
         occ.and(board.rooksqueens[side]);
         attacks.or(occ);
      }
      break;
   case 8:
      //if
      //(!Bitmap::and(board.rooksqueens[side],Bearing::file_mask[f]).is_clear())
      {
         Bitmap occ(board.occupiedR90[White]);
         occ.or(board.occupiedR90[Black]);
         // shift the occupied bits so that the file that sq occupies are
         // in bits 0-7.
         byte atcks = occ.rank_bits(7-f);
         // find the pieces nearest to sq on the file, then
         // "unrotate" the result and shift it back up
         // See if any rooks or queens are there
         attacks.or(Bitmap::and(Bearing::unrot90[Bearing::next_bit_left[atcks][r]][7-f],board.rooksqueens[side]));
      }
      break;
   case -8:
      //if
      //(!Bitmap::and(board.rooksqueens[side],Bearing::file_mask[f]).is_clear())
      {
         Bitmap occ(board.occupiedR90[White]);
         occ.or(board.occupiedR90[Black]);
         // shift the occupied bits so that the file that sq occupies are
         // in bits 0-7.
         byte atcks = occ.rank_bits(7-f);
         // find the pieces nearest to sq on the file, then
         // "unrotate" the result and shift it back up
         // See if any rooks or queens are there
         attacks.or(Bitmap::and(Bearing::unrot90[Bearing::next_bit_right[atcks][r]][7-f],board.rooksqueens[side]));
      }
      break;
   case 7:
      //if (!Bitmap::and(board.bishopsqueens[side],Bearing::diag_a1_mask[attack_square]).is_clear())
      {
         Bitmap occ(board.occupiedDa1[White]);
         occ.or(board.occupiedDa1[Black]);
         DiagInfo *info = &DiagInfos[attack_square][DIAG_LOWER_LEFT];
         // See if any bishops or queens are there
         attacks.or(Bitmap::and(Bearing::unrotda1[Bearing::next_bit_left[
                    occ.rank_bits(info->shift) & info->mask][info->bit]][info->shift],
                                board.bishopsqueens[side]));
      }
      break;
   case -7:
      //if (!Bitmap::and(board.bishopsqueens[side],Bearing::diag_a1_mask[attack_square]).is_clear())
      {
         Bitmap occ(board.occupiedDa1[White]);
         occ.or(board.occupiedDa1[Black]);
         DiagInfo *info = &DiagInfos[attack_square][DIAG_UPPER_RIGHT];
         // See if any bishops or queens are there
         attacks.or(Bitmap::and(Bearing::unrotda1[Bearing::next_bit_right[
                    occ.rank_bits(info->shift) & info->mask][info->bit]][info->shift],
                                board.bishopsqueens[side]));
      }
      break;
   case 9:
      //if (!Bitmap::and(board.bishopsqueens[side],Bearing::diag_a8_mask[attack_square]).is_clear())
      {
         Bitmap occ(board.occupiedDa8[White]);
         occ.or(board.occupiedDa8[Black]);
         DiagInfo *info = &DiagInfos[attack_square][DIAG_LOWER_RIGHT];
         // See if any bishops or queens are there
         attacks.or(Bitmap::and(Bearing::unrotda8[Bearing::next_bit_left[
                    occ.rank_bits(info->shift) & info->mask][info->bit]][info->shift],
                                board.bishopsqueens[side]));
      }
      break;
   case -9:
      //if (!Bitmap::and(board.bishopsqueens[side],Bearing::diag_a8_mask[attack_square]).is_clear())
      {
         Bitmap occ(board.occupiedDa8[White]);
         occ.or(board.occupiedDa8[Black]);
         DiagInfo *info = &DiagInfos[attack_square][DIAG_UPPER_LEFT];
         // See if any bishops or queens are there
         attacks.or(Bitmap::and(Bearing::unrotda8[Bearing::next_bit_right[
                    occ.rank_bits(info->shift) & info->mask][info->bit]][info->shift],
                                board.bishopsqueens[side]));
      }
      break;
   default:  
      break;
   }
}

int attack_estimate( const Board &board, const Move &emove )
{
   ASSERT(!IsNull(emove));
   ColorType my_side = PieceColor(board[StartSquare(emove)]);
   ColorType side = my_side;
   Square square = (TypeOfMove(emove) == EnPassant) ?
                   board.EnPassantSq(OppositeColor(side)) : DestSquare(emove);
   Square attack_square = StartSquare(emove);
   Piece on_square = board[square];
   ASSERT(!IsEmptyPiece(on_square));
   Piece attacker = board[attack_square];
   ASSERT(!IsEmptyPiece(attacker));
   int score_count = 0;
   int score_list[20];
   int swap_score = 0;
   int gain;
   Bitmap opp_attacks(board.calc_attacks(square,OppositeColor(side)));
   if (opp_attacks.is_clear()) // piece is undefended
   {
      if (TypeOfMove(emove) == Promotion)
         return PieceValue(on_square) + PieceValues[PromoteTo(emove)] - 64;
      else
         return PieceValue(on_square);
   }
   Bitmap my_attacks(board.calc_attacks(square,side));
   my_attacks.clear(attack_square); // remove attacking piece
   if (TypeOfPiece(attacker) != Knight)
      addXRay(board,my_attacks,attack_square,square,side);
   int count = 0;
   for (;;)
   {
      attacker = board[attack_square];
      /**
      cout << " " << Piece::Image(attacker.Type())
           << " on " <<
           attack_square.FileImage() <<
           attack_square.RankImage() <<
           " takes " << Piece::Image(on_square.Type())
           << endl;
      **/
      gain = PieceValue(on_square);
      if (TypeOfPiece(attacker) == Pawn &&
          Rank(square,side) == 8)
      {
         if (count == 0) // initial capture is a promotion too
         {
            gain += (PieceValues[PromoteTo(emove)] - 64);
            on_square = MakePiece(PromoteTo(emove),side);
         }
         else
         {
            gain += (576-64); //Piece::Value(Piece::Queen) - Piece::Value(Piece::Pawn);
            on_square = MakePiece(Queen,side);
         }
      }
      else
         on_square = attacker;
      if (side == my_side)
         swap_score += gain;
      else
         swap_score -= gain;
      //if (score_count > 15)
      //    printf("bad score count\n");
      ASSERT(score_count < 20);
      score_list[score_count++] = swap_score;
      side = OppositeColor(side);
      if (side == my_side)
      {
         if (my_attacks.is_clear())
            break;
         attack_square = board.min_attacker(my_attacks,side).first_one();
         if (IsInvalid(attack_square))
            break;
         else
         {
            my_attacks.clear(attack_square);
            addXRay(board,my_attacks,attack_square,square,side);
         }
      }
      else
      {
         if (opp_attacks.is_clear())
            break;
         attack_square = board.min_attacker(opp_attacks,side).first_one();
         if (IsInvalid(attack_square))
            break;
         else
         {
            opp_attacks.clear(attack_square);
            addXRay(board,opp_attacks,attack_square,square,side);
         }
      }
      ++count;
   }
   if (score_count > 1 && score_count % 2 == 0)
   {
      // last capture is by other side, can't avoid it
      score_list[score_count-2] = score_list[score_count-1];
      --score_count; 
   }
   int i = score_count;
   while (i >= 3)
   {
      int max_score = Util::Max(score_list[i-1],score_list[i-2]);
      if (max_score < score_list[i-3])
         score_list[i-3] = max_score;
      i -= 2;
   }
   return score_list[0];
}

