// Copyright 1992-1996 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "bearing.h"
#include "util.h"
#include "beardata.h"

const int RankIncr = 8; // add this to move 1 rank

Bitmap Bearing::knight_attacks[64];
Bitmap Bearing::king_attacks[64];
Bitmap Bearing::pawn_attacks[64][2];
// for each 8-bit value that a rank can have, and each
// offset to a piece within that rank, return the non-zero
// bits closest to the offset bit.
byte Bearing::rank_attacks[256][8];
// for each 8-bit value that a rank can have, and each
// offset to a piece within that rank, return the non-zero
// bits closest to left of the offset bit.
byte Bearing::next_bit_left[256][8];
// for each 8-bit value that a rank can have, and each
// offset to a piece within that rank, return the non-zero
// bits closest to right of the offset bit.
byte Bearing::next_bit_right[256][8];
int Bearing::mobility[256][8];
Bitmap  Bearing::unrot90[256][8];
Bitmap  Bearing::unrotda1[256][8];
Bitmap  Bearing::unrotda8[256][8];
Bitmap  Bearing::file_mask[8];
Bitmap  Bearing::rank_mask[8];
Bitmap  Bearing::diag_a1_mask[64];
Bitmap  Bearing::diag_a8_mask[64];
Bitmap  Bearing::black_squares;
Bitmap  Bearing::white_squares;
int Bearing::diagonal_length[64] = 
{ 1,2,3,4,5,6,7,8,
   2,3,4,5,6,7,8,1,
   3,4,5,6,7,8,1,2,
   4,5,6,7,8,1,2,3,
   5,6,7,8,1,2,3,4,
   6,7,8,1,2,3,4,5,
   7,8,1,2,3,4,5,6,
   8,1,2,3,4,5,6,7};
int Bearing::Directions[64][64];
Bitmap Bearing::rank_file_attack[64];
Bitmap Bearing::diag_attack[64];

static inline void 
SetSquare(
         const Square sq, Square * squares, unsigned &NumSquares)
{
   ASSERT(OnBoard(sq));
   ASSERT(NumSquares < Bearing::MaxBearSq);
   squares[NumSquares++] = sq;
}

static void
PawnMoves(const Board & board,
          const ColorType side, const Square i,
          Square * squares, unsigned &NumSquares)
{
   Square start, dest, dest2;
   Piece piece;

   start = i;
   dest = side == White ? i - RankIncr : i + RankIncr;
   ASSERT(OnBoard(dest));
   if (board[dest] == EmptyPiece)
   {
      // 1-square pawn advance

      SetSquare(dest, squares, NumSquares);
      if (Rank(i,board.Side()) == 2)
      {
         // 2-square pawn advance
         dest2 = (Direction[side] > 0) ?
                 dest + RankIncr : dest - RankIncr;
         if (board[dest2] == EmptyPiece)  // 2-square advance
            SetSquare(dest2, squares, NumSquares);
      }
   }
   // check for possible captures
   if (File(i) != 1)
   {
      dest = (side == White) ? i - (RankIncr + 1) :
             i + (RankIncr - 1);
      piece = board[dest];
      if (piece != EmptyPiece && PieceColor(piece) != board.Side())
         SetSquare(dest, squares, NumSquares);
   }
   if (File(i) != 8)
   {
      dest = (side == White) ? i - (RankIncr - 1) :
             i + (RankIncr + 1);
      piece = board[dest];
      if (piece != EmptyPiece && PieceColor(piece) != board.Side())
         SetSquare(dest, squares, NumSquares);
   }
   Square epsq = board.EnPassantSq(board.OppositeSide());
   if (epsq != InvalidSquare)
   {
      if ((File(i) != 8 && i + 1 == epsq) || 
          (File(i) != 1 && i - 1 == epsq))
      {
         Square dest = epsq + RankIncr * Direction[side];
         if (board[dest] == EmptyPiece)
            SetSquare(dest, squares, NumSquares);
      }
   }
}

static void
PawnAttacks(const Board & board,
            const ColorType side, const Square i,
            Move *moves, unsigned &NumMoves, int ply)
{
   Square dest;
   Piece piece;

   // check for possible captures
   if (File(i) != 1)
   {
      dest = (side == White) ? i - (RankIncr + 1) :
             i + (RankIncr - 1);
      piece = board[dest];
      if (piece != EmptyPiece && (PieceColor(piece) != board.Side()))
      {
         if (Rank(i,side) == 7)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Queen,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Knight,Pawn,TypeOfPiece(piece),Promotion);
            if (ply == 0)
            {
               moves[NumMoves++] =
               CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
               moves[NumMoves++] =
               CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
            }
         }
         else
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,TypeOfPiece(piece),Normal);
      }
   }
   if (File(i) != 8)
   {
      dest = (side == White) ? i - (RankIncr - 1) :
             i + (RankIncr + 1);
      piece = board[dest];
      if (piece != EmptyPiece && (PieceColor(piece) != board.Side()))
      {
         if (Rank(i,side) == 7)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Queen,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Knight,Pawn,TypeOfPiece(piece),Promotion);
            if (ply == 0)
            {
               moves[NumMoves++] =
               CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
               moves[NumMoves++] =
               CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
            }
         }
         else
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,TypeOfPiece(piece),Normal);
      }
   }
   if (Rank(i,side) == 7)
   {
      dest = (side == White) ? i - RankIncr : i + RankIncr;
      piece = board[dest];
      if (piece == EmptyPiece)
      {
         // promotion
         moves[NumMoves++] =
         CreateMove(i,dest,Queen,Pawn,EmptyPiece,Promotion);
         moves[NumMoves++] =
         CreateMove(i,dest,Knight,Pawn,EmptyPiece,Promotion);
         if (ply == 0)
         {
            moves[NumMoves++] =
            CreateMove(i,dest,Rook,Pawn,TypeOfPiece(piece),Promotion);
            moves[NumMoves++] =
            CreateMove(i,dest,Bishop,Pawn,TypeOfPiece(piece),Promotion);
         }
      }
   }
   Square epsq = board.EnPassantSq(board.OppositeSide());
   if (!IsInvalid(epsq))
   {
      ASSERT(TypeOfPiece(board[epsq])==Pawn);
      if ((File(i) != 8 && i + 1 == epsq) || 
          (File(i) != 1 && i - 1 == epsq))
      {
         Square dest = epsq + RankIncr * Direction[side];
         if (board[dest] == EmptyPiece)
            moves[NumMoves++] =
            CreateMove(i,dest,InvalidPiece,Pawn,Pawn,
                       EnPassant);
      }
   }
}

#define RBMoves(board, data, squares, NumSquares) \
    Square i; \
    while ((i = *data) != 255 && board[i] == EmptyPiece) \
    { \
       SetSquare(i,squares,NumSquares); \
       ++data; \
    } \
    if (i != 255 && PieceColor(board[i]) != side) \
       SetSquare(i,squares,NumSquares) \

#define RBAttacks(board, loc, data, moves, NumMoves) \
    Square i; \
    while ((i = *data) != 255 && board[i] == EmptyPiece) \
    { \
       ++data; \
    } \
    if (i != 255 && PieceColor(board[i]) != side) \
       moves[NumMoves++] = \
          CreateMove(loc,i,InvalidPiece,TypeOfPiece(board[loc]), \
          TypeOfPiece(board[i]),Normal) \

unsigned 
Bearing::BearSq(const Board & board,
                const Square loc, Square * squares)
{
   unsigned NumSquares = 0;
   int j;

   const ColorType side = PieceColor(board[loc]);
   switch (TypeOfPiece(board[loc]))
   {
   case EmptyPiece:
      break;
   case Pawn:
      PawnMoves(board, side,
                loc, squares, NumSquares);
      break;
   case Knight:
      {
         const ColorType side = PieceColor(board[loc]);
         const byte *data = KnightSquares[(int)loc];
         int i = 0;
         while (i < 8 && *data != 255)
         {
            ++i;
            Square dest(*data++);
            Piece piece = board[dest];
            if (piece == EmptyPiece || PieceColor(piece) != side)
               SetSquare(dest, squares, NumSquares);
         }
      }
      break;
   case Bishop:
      for (j = 0; j < 4; j++)
      {
         const byte *data = BishopSquares[loc] + (j*8);
         RBMoves(board, data, squares, NumSquares);
      }
      break;
   case Rook:
      for (j = 0; j < 4; j++)
      {
         const byte *data = RookSquares[loc] + (j*8);
         RBMoves(board, data, squares, NumSquares);
      }
      break;
   case Queen:
      for (j = 0; j < 4; j++)
      {
         const byte *data = BishopSquares[loc] + (j*8);
         RBMoves(board, data, squares, NumSquares);
      }
      for (j = 0; j < 4; j++)
      {
         const byte *data = RookSquares[loc] + (j*8);
         RBMoves(board, data, squares, NumSquares);
      }
      break;
   case King:
      {
         const byte *data = KingSquares[(int)loc];
         Bitmap king_atcks(king_attacks[board.KingPos(
          board.OppositeSide())]);
         for (int i = 0; i <8 && *data != 255 ;i++)
         {
            Square dest(*data++);
            Piece piece = board[dest];
            if (piece == EmptyPiece || PieceColor(piece) != side)
            {
               if (!king_atcks.is_set(dest))
                 SetSquare(dest, squares, NumSquares);
            }
         }
      }
   case InvalidPiece:
      break;
   }
   return NumSquares;
}

unsigned 
Bearing::Captures(const Board & board,
                  const Square loc, Move *moves, int ply)
{
   unsigned NumMoves = 0;
   int j;

   const ColorType side = PieceColor(board[loc]);
   switch (TypeOfPiece(board[loc]))
   {
   case EmptyPiece:
      break;
   case Pawn:
      PawnAttacks(board, side,
                  loc, moves, NumMoves, ply);
      break;
   case Knight:
      {
         Bitmap dests(knight_attacks[loc]);
         if (side == White)
            dests.and(board.occupied[Black]);
         else
            dests.and(board.occupied[White]);
         Bitmap it(dests);
         Square sq;
         for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
         {
            moves[NumMoves++] = 
            CreateMove(loc,sq,InvalidPiece,Knight,
                       TypeOfPiece(board[sq]),Normal);
         }
      }
      break;
   case Bishop:
      for (j = 0; j < 4; j++)
      {
         const byte *data = BishopSquares[loc] + (j*8);
         RBAttacks(board, loc, data, moves, NumMoves);
      }
      break;
   case Rook:
      for (j = 0; j < 4; j++)
      {
         const byte *data = RookSquares[loc] + (j*8);
         RBAttacks(board, loc, data, moves, NumMoves);
      }
      break;
   case Queen:
      for (j = 0; j < 4; j++)
      {
         const byte *data = BishopSquares[loc] + (j*8);
         RBAttacks(board, loc, data, moves, NumMoves);
      }
      for (j = 0; j < 4; j++)
      {
         const byte *data = RookSquares[loc] + (j*8);
         RBAttacks(board, loc, data, moves, NumMoves);
      }
      break;
   case King:
      {
         Bitmap dests(king_attacks[loc]);
         if (side == White)
            dests.and(board.occupied[Black]);
         else
            dests.and(board.occupied[White]);
         dests.andNot(king_attacks[board.KingPos(
          board.OppositeSide())]);
         Bitmap it(dests);
         Square sq;
         for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
         {
            moves[NumMoves++] = 
            CreateMove(loc,sq,InvalidPiece,King,
                       TypeOfPiece(board[sq]),Normal);
         }
      }
      break;
   case InvalidPiece:
      break;
   }
   return NumMoves;
}

unsigned 
Bearing::Attack(const Board & board,
                const Square loc, const ColorType side,
                Square * squares)
{
   ASSERT(OnBoard(loc));
   int index = 0;
   int i, n;
   Square square1, sq;
   Piece my_pawn = MakePiece(Pawn, side);
   if (side == Black)
   {
      if (board[loc] == EmptyPiece)
      {
         Square origin = loc-RankIncr;
         if (OnBoard(origin) &&
             board[origin] == my_pawn)
            squares[index++] = origin;
         if (Rank(loc,side) == 4 && board[origin] == EmptyPiece &&
             board[origin - RankIncr] == my_pawn)
            squares[index++] = origin - RankIncr;
      }
      else if (Rank(loc,side) != 1)
      {
         // square is occupied, try pawn captures
         if (File(loc) != 1 &&
             board[loc - (RankIncr+1)] == my_pawn)
            squares[index++] = loc - (RankIncr+1);
         if (File(loc) != 8 &&
             board[loc - (RankIncr-1)] == my_pawn)
            squares[index++] = loc - (RankIncr-1);
      }
   }
   else
   {
      if (board[loc] == EmptyPiece)
      {
         Square origin = loc+RankIncr;
         if (OnBoard(origin) &&
             board[origin] == my_pawn)
            squares[index++] = origin;
         if (Rank(loc,side) == 4 && board[origin] == EmptyPiece &&
             board[origin + RankIncr] == my_pawn)
            squares[index++] = origin + RankIncr;
      }
      else if (Rank(loc,side) != 1)
      {
         if (File(loc) != 8 &&
             board[loc + RankIncr+1] == my_pawn)
            squares[index++] = loc + RankIncr + 1;
         if (File(loc) != 1 &&
             board[loc + RankIncr-1] == my_pawn)
            squares[index++] = loc + RankIncr - 1;
      }
   }
   Piece knight = MakePiece(Knight, side);
   Piece king = MakePiece(King, side);
   for (i = 0; i < 8; i++)
   {
      sq = KnightSquares[loc][i];
      if (sq == 255)
         break;
      if (board[sq] == knight)
         squares[index++] = sq;
   }
   for (i = 0; i <8; i++)
   {
      sq = KingSquares[loc][i];
      if (sq == 255)
         break;
      if (board[sq] == king)
         squares[index++] = sq;
   }
   n = index;
   Piece bishop = MakePiece(Bishop, side);
   Piece queen = MakePiece(Queen, side);
   Piece rook = MakePiece(Rook, side);
   BOOL found;
   for (i = 0; i < 4; i++)
   {
      square1 = loc;
      found = FALSE;
      const byte *data = BishopSquares[loc] + (8*i);
      while (*data != 255)
      {
         Piece piece = EmptyPiece;
         do
         {
            if (*data == 255) break;
            square1 = *data++;
            piece = board[square1];
         } while (piece == EmptyPiece);
         if (piece == bishop || piece == queen)
         {
            found = TRUE;
            ++n;
            squares[index] = square1;
            break;
         }
         else
            break;
      }
      if (found)
         ++index;
      data = RookSquares[loc] + (8*i);
      square1 = loc;
      found = FALSE;
      for (;;)
      {
         Piece piece = EmptyPiece;
         do
         {
            if (*data == 255)
               break;
            square1 = *data++;
            piece = board[square1];
         } while (piece == EmptyPiece);
         if (piece == rook || piece == queen)
         {
            found = TRUE;
            ++n;
            squares[index] = square1;
            break;
         }
         else
            break;
      }
      if (found)
         ++index;
   }
   ASSERT(n < 36);
   return n;
}

#pragma optimize("g",off)
BOOL Bearing::Pinned( const Board &board, const Square loc,
                      Piece &PinnedByPiece, 
                      Square &PinnedBySquare,
                      int &dir)
{
   Piece p = board[loc];
   if (p == EmptyPiece || TypeOfPiece(p) == King)
      return FALSE;
   const ColorType side = ColorOfPiece(p);
   const ColorType oside = OppositeColor(side);
   // We "cheat" on constness by altering board[loc].  But we will
   // put the original contents back .. honest
   Piece &place = (Piece &)board[loc];
   place = EmptyPiece; // imagine me gone
   Square ks = board.KingPos(side);
   int offset = (int)loc - (int)ks;
   int signOfOffset = Util::Sign(offset);
   BOOL attackerFound = FALSE;
   if (File(ks) == File(loc))  // possible pin on file
   {
      const byte *data = (signOfOffset > 0) ?
                         RookSquares[loc]+8 : RookSquares[loc];
      Square newsquare;
      Piece attackPiece;
      while (*data != 255)
      {
         newsquare = *data++;
         attackPiece = board[newsquare];
         if (attackPiece != EmptyPiece)
            break;
      }
      if (PieceColor(attackPiece) == oside && (
                                              (TypeOfPiece(attackPiece) == Rook) ||
                                              (TypeOfPiece(attackPiece) == Queen)))
      {
         place = p; // put piece back
         PinnedBySquare = newsquare;
         PinnedByPiece = attackPiece;
         dir = ((int)newsquare - (int)loc > 0) ? RankIncr : -RankIncr;
         attackerFound = TRUE;
      }
   }
   else if (Rank(ks,White) == Rank(loc,White))  // possible pin on rank
   {
      const byte *data = (signOfOffset > 0) ?
                         RookSquares[loc]+16 : RookSquares[loc]+24;
      Square newsquare;
      Piece attackPiece;
      while (*data != 255)
      {
         newsquare = *data++;
         attackPiece = board[newsquare];
         if (attackPiece != EmptyPiece)
            break;
      }
      if (PieceColor(attackPiece) == oside && (
                                              (TypeOfPiece(attackPiece) == Rook) ||
                                              (TypeOfPiece(attackPiece) == Queen)))
      {
         place = p; // put piece back
         PinnedBySquare = newsquare;
         PinnedByPiece = attackPiece;
         dir = ((int)newsquare > (int)loc) ? 1 : -1;
         attackerFound = TRUE;
      }
   }
   else if (SquareColor(ks) == SquareColor(loc))
   {
      if (offset % 9 == 0) // possible diagonal pin
      {
         const byte *data = (signOfOffset > 0) ?
                            BishopSquares[loc]+16 : BishopSquares[loc]+8;
         Square newsquare;
         Piece attackPiece;
         while (*data != 255)
         {
            newsquare = *data++;
            attackPiece = board[newsquare];
            if (attackPiece != EmptyPiece)
               break;
         }
         if (PieceColor(attackPiece) == oside && (
                                                 (TypeOfPiece(attackPiece) == Bishop) ||
                                                 (TypeOfPiece(attackPiece) == Queen)))
         {
            place = p; // put piece back
            PinnedBySquare = newsquare;
            PinnedByPiece = attackPiece;
            dir = ((int)newsquare > (int)loc) ? 9 : -9;
            attackerFound = TRUE;
         }
      }
      else if (offset % 7 == 0) // possible diagonal pin
      {
         const byte *data = (signOfOffset > 0) ?
                            BishopSquares[loc] + 24 : BishopSquares[loc];
         Square newsquare;
         Piece attackPiece;
         while (*data != 255)
         {
            newsquare = *data++;
            attackPiece = board[newsquare];
            if (attackPiece != EmptyPiece)
               break;
         }
         if (PieceColor(attackPiece) == oside && (
                                                 (TypeOfPiece(attackPiece) == Bishop) ||
                                                 (TypeOfPiece(attackPiece) == Queen)))
         {
            place = p; // put piece back
            PinnedBySquare = newsquare;
            PinnedByPiece = attackPiece;
            dir = ((int)newsquare > (int)loc) ? 7 : -7;
            attackerFound = TRUE;
         }
      }
   }
   place = p;
   if (attackerFound)
   {
      // see if the path is clear to the king
      Square sq(loc);
      do
      {
         sq -= dir;
         if (!OnBoard(sq) || (sq != ks && board[sq] != EmptyPiece))
            return FALSE;
      } while (sq != ks);
      return TRUE;
   }
   else
      return FALSE;
}

void Bearing::init()
{
   for (int i = 0; i < 64; i++)
   {
      Square sq = i;
      file_mask[File(sq)-1].set(i);
      rank_mask[Rank(sq,Black)-1].set(i);
      if (SquareColor(sq) == White)
         white_squares.set(i);
      else
         black_squares.set(i);
      const byte *data = KnightSquares[i];
      int j = 0;
      while (j < 8 && *data != 255)
      {
         knight_attacks[i].set(*data++);
         ++j;
      }
      const byte *data2= KingSquares[i];
      j = 0;
      while (j < 8 && *data2 != 255)
      {
         king_attacks[i].set(*data2++);
         ++j;
      }

      if (File(sq) != 1)
      {
         if (i-9>=0) pawn_attacks[i][Black].set(i-9);
         if (i+7<=63) pawn_attacks[i][White].set(i+7);
      }
      if (File(sq) != 8)
      {
         if (i-7>=0) pawn_attacks[i][Black].set(i-7);
         if (i+9<=63) pawn_attacks[i][White].set(i+9);
      }
      Square sq2 = sq;
      diag_a1_mask[sq].set(sq);
      diag_a8_mask[sq].set(sq);
      while (sq2 >= 7 && File(sq2) != 8)
      {
         sq2 = sq2 - 7;
         diag_a1_mask[sq].set(sq2);
      }
      sq2 = sq;
      while (sq2 <= 63-7 && File(sq2) != 1)
      {
         sq2 = sq2 + 7;
         diag_a1_mask[sq].set(sq2);
      }
      sq2 = sq;
      while (sq2 >= 9 && File(sq2) != 1)
      {
         sq2 = sq2 - 9;
         diag_a8_mask[sq].set(sq2);
      }
      sq2 = sq;
      while (sq2 <= 63-9 && File(sq2) != 8)
      {
         sq2 = sq2 + 9;
         diag_a8_mask[sq].set(sq2);
      }
      for (j = 0; j < 64; j++)
      {
         if (i == j)
         {
            Directions[i][j] = 0;
            continue;
         }
         Square sq1 = i;
         Square sq2 = j;
         int offset;
         int abs = (sq2 > sq1) ? (int)sq2 - (int)sq1 : (int)sq1 - (int)sq2;
         if (File(sq1) == File(sq2))
            offset = RankIncr;
         else if (Rank(sq1,White) == Rank(sq2,White))
            offset = 1;
         else if (SquareColor(sq1) == SquareColor(sq2))
         {
            if (abs % (RankIncr+1) == 0)
               offset = RankIncr+1;
            else if (abs % (RankIncr-1) == 0)
               offset = RankIncr-1;
            else
            {
               Directions[i][j] = 0;
               continue;
            }
         }
         else
         {
            Directions[i][j] = 0;
            continue;
         }
         Directions[i][j] = (sq2 > sq1) ? offset : -offset;
         if (File(i) == File(j) ||
             Rank(i,White) == Rank(j,White))
            rank_file_attack[i].set(j);
      }
   }
   for (i=0; i<64; i++)
      for (int j=0;j<64;j++)
         if (Directions[i][j]==9 ||
             Directions[i][j]==-9 ||
             Directions[i][j]==7 ||
             Directions[i][j]==-7)
            diag_attack[i].set(j);
   int m;
   for (i=0; i<256; i++)
   {
      for (int off=0; off <8; off++)
      {
         byte b,lb,rb;
         b = lb = rb = 0;
         m = 0;
         for (int j=off+1;j<8;j++)
         {
            if (i & (1<<j))
            {
               b |= (1<<j);
               rb |= (1<<j);
               break;
            }
            ++m;
         }
         for (j=off-1;j>=0;--j)
         {
            if (i & (1<<j))
            {
               b |=(1<<j);
               lb |= (1<<j);
               break;
            }
            ++m;
         }
         rank_attacks[i][off] = b;
         next_bit_left[i][off] = lb;
         next_bit_right[i][off] = rb;
         mobility[i][off] = m;
         // compute a bitmap used for unrotating 90-degree bitboards
         Bitmap bm(i);
         Bitmap bunrot;
         for (int k=0;k<8;k++)
         {
            if (bm.is_set(k))
               bunrot.set(Bitmap::unrot90[k+8*off]);
         }
         unrot90[i][off] = bunrot;
         Bitmap bunrot2;
         for (k=0;k<8;k++)
         {
            if (bm.is_set(k))
            {
               bunrot2.set(Bitmap::unrot_da1[k+8*off]);
            }
         }
         unrotda1[i][off] = bunrot2;
         Bitmap bunrot3;
         for (k=0;k<8;k++)
         {
            if (bm.is_set(k))
            {
               bunrot3.set(Bitmap::unrot_da8[k+8*off]);
            }
         }
         unrotda8[i][off] = bunrot3;

      }
   }
}
