// Copyright 1996 by Jon Dart.  All Rights Reserved.
#ifndef _BITMAP_H
#define _BITMAP_H

#include "types.h"
#include "square.h"
#include <iostream.h>

#define FAST_SHIFT_R(data,n) \
    (n>=32) ? Bitmap(data.hivalue()>>(n)-32) \
         : Bitmap((uint32)data.value() >> (n))
         
class Bitmap
{
       friend ostream & operator << (ostream &o, const Bitmap &b);

    public:	

       struct ints
       {
          uint32 loval,hival;
       };
       struct shorts
       {
          uint16 lo1,lo2,hi1,hi2;
       };
       union conv
       {
          uint64 val1;
          ints val2;
          shorts val3;
       };

    Bitmap()
    : data(0)
    {
    }
    
    Bitmap(uint64 n)
    : data(n)
    {
    }
    
    uint64 value() const
    {
       return data;
    }
    
    void *addr() const
    {
       return (void*)&data;
    }
    
    byte byte_value() const
    {
       // PC is little-endian
       return *((byte*)&data);
    }
    
    uint32 hivalue() const
    {
       return ((conv*)(this))->val2.hival;
    }
    
    uint32 lovalue() const
    {
       return ((conv*)(this))->val2.loval;
    }
    
    void clear()
    {
       data = 0;
    }
    void set_lsb()
    {
     data &= ~(data-1);
    }
    void set(int n)
    {
       data |= all_masks[n].mask;
    }
    void clear(int n)
    {
       data &= all_masks[n].not_mask;
    }
    void setR90(int n)
    {
        data |= all_masks[n].maskR90;
    }
    void clearR90(int n)
    {
        data &= all_masks[n].not_maskR90;
    }
    void setDa1(int n)
    {
        data |= all_masks[n].maskDa1;
    }
    void clearDa1(int n)
    {
        data &= all_masks[n].not_maskDa1;
    }
    void setDa8(int n)
    {
        data |= all_masks[n].maskDa8;
    }
    void clearDa8(int n)
    {
        data &= all_masks[n].not_maskDa8;
    }
    void setClear(const Bitmap &b)
    {
        data ^= b.data;
    }
    int is_set(int n) const
    {
        return ((data & mask[n]) != (uint64)0);
    }
    int is_clear()const
    {
        return (data == (uint64)0);
    }
    static Bitmap and( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data & src2.data);
    }
    void and( const Bitmap &b)
    {
        data &= b.data;
    }
    void andNot(const Bitmap &b)
    {
       data &= ~b.data;
    }
    static Bitmap andNot( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data & ~src2.data);
    }
    void orNot(const Bitmap &b)
    {
       data |= ~b.data;
    }
    static Bitmap orNot( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data | ~src2.data);
    }
    static Bitmap or( const Bitmap &src1, const Bitmap &src2)
    {
       return Bitmap(src1.data | src2.data);
    }
    void or( const Bitmap &b)
    {
        data |= b.data;
    }
    static Bitmap shr(const Bitmap &src, int n)
    {
        return Bitmap(src.data >> n);
    }
    void shr(int n)
    {
        data = data >> n;    
    }
    static Bitmap shl(const Bitmap &src, int n)
    {
        return Bitmap(src.data << n);
    }
    void shl(int n)
    {
        data = data << n;    
    }
    void shl_byte(int n)
    {
        // shift lowest 8 bits left by n bytes
        char *p = (char*)&data;
        p[n] = *p;
        if (n>0) *p = 0;
    }
    static Bitmap Bitmap::shl_byte(byte b, int n)
    {
        // shift lowest 8 bits left by n bytes
        uint64 new_data = 0;
        char *p = (char*)&new_data;
        p[n] = b;
        return Bitmap(new_data);
    }
    
    int operator == (const Bitmap &b)
    {
       return b.data == data;
    }

    int operator != (const Bitmap &b)
    {
       return b.data != data;
    }

    // extract the bits for the nth rank (from the Black side)
    unsigned rank_bits(int n) const
    {
       byte *p = (byte*)&data;
       return (unsigned)p[n];
    }

    // return the number of bits set
    unsigned int bit_count() const
    {
       unsigned register int n;
       uint64 tmp = data;
       for (n = 0; tmp != 0; n++, tmp &= (tmp - 1));
       return n;
    }
    
    Square first_one() const
    {
      if (data == 0) return InvalidSquare;
      extern int *first_ones;
      register uint32 copy;
      int word = (lovalue() == 0);
      copy = ((uint32*)&data)[word];
      copy &= ~(copy-1);
      if (copy & 0xffff)
           return first_ones[copy & 0xffff] + word*32;
      else 
           return first_ones[copy >> 16]+word*32+16;
/**
      extern int *first_ones_8bit;
      register uint32 copy;
      int word = (lovalue() == 0);
      copy = ((uint32*)&data)[word];
      copy &= ~(copy-1);
      if (copy & 0xffff)
      {
           if (copy & 0xff00)
              return first_ones_8bit[copy >>8] + word*32 + 8;
           else
              return first_ones_8bit[copy] + word*32;
      }
      else 
      {
           copy = copy >> 16;
           if (copy & 0xff00)
              return first_ones_8bit[copy >>8] + word*32 + 24;
           else
              return first_ones_8bit[copy] + word*32 + 16;
      }
**/
    }
    
    static void init();
    
    static void cleanup();

    static const uint64 mask[64];
    static int unrot90[64];
    static int R90[64];
    static int da1[64];
    static int da1mod8[64];
    static int da1div8[64];
    static int unrot_da1[64];
    static int da1_mask[64];
    static int da8[64];
    static int da8mod8[64];
    static int da8div8[64];
    static int unrot_da8[64];
    static int da8_mask[64];
    
    struct masks
    {
       uint64 mask;
       uint64 maskDa1;
       uint64 maskDa8;
       uint64 maskR90;
       uint64 not_mask;
       uint64 not_maskDa1;
       uint64 not_maskDa8;
       uint64 not_maskR90;
    };
    
    static masks all_masks[64];

    uint64 data;

};

inline uint64 and( const Bitmap &src1, const Bitmap &src2)
{
    return (src1.data & src2.data);
}

inline uint64 or( const Bitmap &src1, const Bitmap &src2)
{
    return (src1.data | src2.data);
}

#define DIAG_UPPER_LEFT 0
#define DIAG_UPPER_RIGHT 1
#define DIAG_LOWER_LEFT 2
#define DIAG_LOWER_RIGHT 3

struct DiagInfo
{
   int shift;
   int bit;
   int mask;
};

extern DiagInfo DiagInfos[64][4];


#endif
