// Copyright 1997 by Jon Dart.  All Rights Reserved.

#ifndef _BOARD_STATE_H
#define _BOARD_STATE_H

#include "board.h"

// This class maintains state information about a board
// position that is needed to restore the board state
// when a move is "undone".
//
class Board_State
{
   friend class Board;
public:
   Board_State(const Board &b)
   {
      old_hashcode = b.HashCode();
      old_pawn_hashcode = b.pawn_hash();
      old_castlestatus[White] = b.CastleStatus(White);
      old_castlestatus[Black] = b.CastleStatus(Black);
      old_enpassantsq[White] = b.EnPassantSq(White);
      old_enpassantsq[Black] = b.EnPassantSq(Black);
      old_checkStatus = b.CheckStatus();
      old_movecount = b.move_count;
   }
   
private:   
	hash_t old_hashcode, old_pawn_hashcode;
        int old_movecount;
	Board::CastleType old_castlestatus[2];
	Square old_enpassantsq[2];
        Board::CheckStatusType old_checkStatus;
};

#endif

