// Copyright 1997, 1998, 1999 by Jon Dart. All Rights Reserved.
#include "calctime.h"
#include "globals.h"
#include "util.h"

// Calculate time limit for a 0-increment game
static int zero_inc_time_limit(int minutes,int time_left)
{
    int time_target;
    if (time_left < 3000) // 30 seconds
       time_target = (time_left+1000)/80;
    else if (time_left < 1500*minutes) // < 1/4 time left
       time_target = (100*(time_left+1000))/3000;
    else if (time_left < 3000*minutes) // < 1/2 time left
       time_target = (100*(time_left+4000))/3000;
    else
       time_target = 120*(minutes > 3 ? minutes + 1 : minutes);
    return time_target;
}

DWORD calc_time_limit(int moves, int minutes, int incr, 
   int32 time_left, int32 opp_time)
{
        int time_target = 0;
        int moves_in_game = game_moves->num_moves()/2; // full moves, not half-moves
        int est_time = incr/4 + minutes;
        int pct_time_left = (time_left*100)/(6000*minutes);
        if (incr > 0 && minutes == 0)
        {
            // Fixed time per move
            if (time_left > 400*incr)
            {
               time_target = incr*150;
            }
            else if (time_left > 250*incr)
            {
               time_target = incr*125;
            }
            else if (time_left > 125*incr)
            {
               time_target = 80*incr;
            }
            else
            {
               time_target = Util::Min(time_left/2,incr*60);
            }
        }
        else if (incr >0)
        {
            // Fischer-type time control

            int extra_time = (time_left - 6000*minutes);
            if (extra_time < 0) extra_time = 0;
            
            time_target = zero_inc_time_limit(est_time,time_left) 
             + (extra_time/12);
            if (time_left < incr*100)
            {
               time_target = 0;
            }
            else if (time_left < 6000)
            {
               time_target = incr*75 + (time_target*time_left)/6000;
            }
            else
               time_target = incr*90 + (pct_time_left*incr)/2 + time_target;

            // try not to fall behind the opponent
            if (pct_time_left < 75) // less than 3/4 initial value
            {
                if (time_left > 3000)
                {
                   int threshold = Util::Min(9000,500*est_time);
                   int max_boost = time_target/4;
                   if (opp_time > time_left + threshold)
                   {
                       time_target -= Util::Min(max_boost,(opp_time-time_left)/10);
                       if (time_target < incr*100)
                            time_target = 90*incr;
                   }
                   else if (time_left > opp_time + threshold)
                   {
                       time_target += Util::Min(max_boost,(time_left-opp_time)/10);
                   }
                }
            }
        }
        else // zero-increment or "Tournament" game
        {
            if (time_left > 500)
                time_left-= 500;
            if (moves == 0) // "Game" time limit
            {
               time_target = zero_inc_time_limit(minutes,time_left);
               // try not to fall behind the opponent
               int threshold = Util::Min(9000,500*est_time);
               int max_boost = time_target/4;
               if (opp_time > time_left + threshold && time_left > 3000)
               {
                   time_target -= Util::Min(max_boost,(opp_time-time_left)/10);
               }
               else if (time_left > opp_time + threshold)
               {
                   time_target += Util::Min(max_boost,(time_left-opp_time)/10);
               }
            }
            else
            {
               int moves_to_time_control = moves-(moves_in_game % moves);
               time_target = time_left/moves_to_time_control;
               int slack = Util::Max(500,time_target/10);
               if (time_left > slack+100)
                  time_target = (time_left-slack)/moves_to_time_control; 
               else
                  time_target = 0;
            }
        }
        // set a minimum (0.3 sec).
        if (time_target < 30) time_target = 30;
//        cout << "(debug) time target = " << time_target << endl;
        if (time_target > time_left) 
        {
           cout << "(debug) warning : time_target > time_left" << endl;
           if (incr)
              time_target = 50*incr;
           else
             time_target = time_left/4;
        }
        return time_target;
}

