// Copyright 1994 by Jon Dart.  All Rights Reserved.
#ifndef __CHESSIO_H__
#define __CHESSIO_H__

#include "board.h"
#include "epdrec.h"
#include <fstream.h>

class Log;

class ChessIO
{
public:

    /**** PGN file support ****/
    
    // scan a PGN file, fill "contents" with strings.  Each string contains
    // a descriptive 1-line header followed by a tab and the file offset
    // of the game in decimal digits. 
    static BOOL scan_pgn(ifstream &ifs, CStringList &contents);

#ifndef _CONSOLE
    // read in a game in PGN format.
    static BOOL load_pgn(Board &board, ifstream &ifs, Log &log, DWORD
    offset,
     CPtrArray *& headers);
#endif
    static BOOL store_pgn(ostream &ofs, Log &log,
                          const ColorType computer_side,
                          CPtrArray &headers);

    static void get_game_description(CPtrArray &hdrs, CString &descr);

    // read & write positions in FEN format.
    static BOOL load_fen(istream &ifs, Board &board);
    static BOOL store_fen(ofstream &ofs, const Board &board);

    // read a PGN game header from the file.
    static void collect_headers(ifstream &game_file, CPtrArray *&hdrs, long &first);

    enum Token { Number, GameMove, Result, Comment, Unknown, Ignore, Eof };
    // read the next token from the "body" of a PGN game.
    static Token get_next_token(ifstream &game_file, char *buf, int limit);

    static EPD_Record * read_epd_record(istream &ifs, Board &board);
#ifdef _WINDOWS
    /**** EPD file support ****/
    
    static void execute_epd_record(EPD_Record *rec, Board &board);
    static void annotate_epd_record(EPD_Record *rec, Board &board,
                                    ofstream &outfile);
#endif    
    
};

#endif
