// Copyright 1994 by Jon Dart.  All Rights Reserved.

#ifndef _MATERIAL_H
#define _MATERIAL_H

#include "types.h"
#include "piece.h"

class Material
{
	friend class Board;
public:
	Material():
        info(0), total(0), level(0)
	{
	}
	
        enum {KP = 0x100001};
        enum {KBP = 0x100101};
        enum {KNB = 0x100110};
        enum {KNN = 0x100020};
        enum {KB  = 0x100100};
        enum {KN  = 0x100010};
        
        void add_piece(const PieceType p)
        {
           info += masks[(int)p];
           total += PieceValue(p);
           level += levels[(int)p];
           ASSERT(level >=0);
        }

	void remove_piece(const PieceType p)
        {
          info -= masks[(int)p];
          total -= PieceValue(p);
          level -= levels[(int)p];
          ASSERT(level >=0);
        }

	// return the number of pieces of the given type:
	unsigned count(const PieceType) const;

	// return the total material value:
	uint32 value() const
	{
		return total;
	}
	
	uint32 infobits() const
	{
		return info;
	}
	
	// return the total number of pieces (excluding pawns and king)
	unsigned piece_count() const;
        
        // return the number of pawns
        unsigned pawn_count() const
        {
                return (info & 0xf);
        }
        
        unsigned queen_count() const
        {
                return (info & 0xf0000) >> 16;
        }

        BOOL has_queen() const
        {
                return (info & 0xf0000);
        }

        unsigned rook_count() const
        {
                return (info & 0xf000) >> 12;
        }

        unsigned knight_count() const
        {
                return (info & 0xf0) >> 4;
        }

        unsigned bishop_count() const
        {
                return (info & 0xf00) >> 8;
        }

	// True if only king and pawns:
	BOOL no_pieces() const
	{
		return (info & (0xffffffff-0x100000-0xf)) == 0;
	}
        
        // True if no pawns
        BOOL no_pawns() const
        {
                return (info & 0xf) == 0;
        }
	
	// True if bare king:
	BOOL king_only() const
	{
		return (info & 0xfffff) == 0;
	}
        
        // This provides a rough measure of the material
        // (used mostly for king safety calculation)
        int material_level() const
        {
           return level;
        }
	
private:
	void clear();

	uint32 total;
	int32 info;
        int32 level;
        static const int32 masks[8];
        static const int32 masks2[8];
        static const int32 levels[8];
};

#endif

