// Copyright 1994 by Jon Dart.  All Rights Reserved

#include "stdafx.h"
#include "board.h"
#include "util.h"
#include <iostream.h>
#include <iomanip.h>
extern "C"
{
#include <ctype.h>
#include <string.h>
};

static SpecialType get_move_type(const Board &board, Move m)
{
    // set special field
    SpecialType my_special = Normal;
    switch (PieceMoved(m))
    {
    case Pawn:
       if (Rank(DestSquare(m),board.Side()) == 8)
       {
          my_special = Promotion;
       }
       else if (IsEmptyPiece(board[DestSquare(m)]))
       {
          if (File(DestSquare(m)) != File(StartSquare(m)))
	  {
	     my_special = EnPassant;
	  }
       }
       break;
    case King:
       if (Util::Abs((int)StartSquare(m) - (int)DestSquare(m)) == 2)
       {
          if (StartSquare(m) > DestSquare(m))
	         my_special = QCastle;
          else
	         my_special = KCastle;
       }
       break;
    default:
       break;
    }
    return my_special;
}

Move CreateMove(const Board &board, Square start, Square dest, PieceType promotion )
{
   MoveUnion mu;
   mu.contents.my_start = start;
   mu.contents.my_dest = dest;
   mu.contents.my_promotion = (byte)promotion;
   mu.contents.flag = NewMove;
   if (IsInvalid(start))
   {
        return NullMove;
   }
   else
   {
      mu.contents.pieceMoved = TypeOfPiece(board[start]);
      ASSERT(mu.contents.pieceMoved != EmptyPiece);
   }
   mu.contents.capture = TypeOfPiece(board[dest]);
   mu.contents.type = get_move_type(board,mu.m);
   if (mu.contents.type == EnPassant)
      mu.contents.capture = Pawn;
   mu.contents.flag = NewMove;
   return mu.m;
}	
const char * MoveImage(Move m) 
{
    static char image[8];
    if (IsNull(m))
    {
       strcpy(image,"(null)");
       return image;
    }
    image[0] = FileImage(StartSquare(m));
    image[1] = RankImage(StartSquare(m));
    if (Capture(m) != EmptyPiece)
      image[2] = 'x';
    else
      image[2] = '-';
    image[3] = FileImage(DestSquare(m));
    image[4] = RankImage(DestSquare(m));
    int i = 5;
    if (PromoteTo(m) != EmptyPiece && PromoteTo(m) != InvalidPiece)
    {
       image[i++] = '=';
       image[i++] = PieceImage(PromoteTo(m));
    }
    image[i] = '\0';
    return image;
}


#ifdef _TRACE
ostream & operator << (ostream &o, Move &move)
{
	o << MoveImage(move);
	return o;
}
#endif
