// Copyright 1993, 1995 by Jon Dart. All Rights Reserved.

#include "stdafx.h"
#include "pinfo.h"

Position_Info::Position_Info()
{
}

Position_Info::Position_Info( const Board &board, int rep_count )
: my_hashcode(board.HashCode())
{
    
    uint16 x = (uint16)(board.CastleStatus(Black));
    uint16 y = (uint16)(board.CastleStatus(White));
    uint16 z = (uint16)(board.EnPassantSq(board.OppositeSide())) & 0x3f;
    my_xhashcode = x | (y << 3) | (z << 8) | (rep_count << 6);
}

Position_Info::Position_Info( const Board &board, const int depth,
                              const byte age,
	           	      const ValueType hashflags,
			      const BOOL forced,
                              const BOOL in_check,
			      const int rep_count,
			      const int value,
			      const Move &best_move)
:  my_depth(depth),
   my_age(age),
   my_hashcode(board.HashCode()),
   my_value(value),start(StartSquare(best_move)),
   dest(DestSquare(best_move)),
   promotion((byte)PromoteTo(best_move))
{
    my_flags = (byte)hashflags;
    if (forced)
        my_flags |= 0x80;
    if (in_check)
        my_flags |= 0x40;
    ASSERT(rep_count <= 3);
    // The "xhashcode" byte holds info on castling and en passant
    // status.
    uint16 x = (uint16)(board.CastleStatus(Black));
    uint16 y = (uint16)(board.CastleStatus(White));
    uint16 z = (uint16)(board.EnPassantSq(board.OppositeSide())) & 0x3f;
#ifdef _DEBUG
    if (TypeOfMove(best_move) == EnPassant)
       ASSERT(z != 63);
#endif       
    my_xhashcode = x | (y << 3) | (z << 8) | (rep_count << 6);
}

Move Position_Info::best_move(const Board &b) const
{
    if (start == InvalidSquare)
       return NullMove;
    else
       return CreateMove(b,(int)start,(int)dest,(PieceType)promotion);
}

