// Copyright 1992 by Jon Dart.  All Rights Reserved.

#ifndef _SCORING_H
#define _SCORING_H

#include "board.h"
class King_Pawn_Entry;

class Scoring
{
	// This class does static evaluation of chess positions.

	public:
		
	static void init(const Board &board);
        
        static void cleanup();
		
	static int evalu8( const Board &board );
	// evaluates "board" from the perspective of the side to move.
		
	static int positional_score( const Board &board, int alpha, int beta );
	// returns a positional score

	static int material_score( const Board &board );
	// returns a material score
		
	static int material_score( const Board &board, Move m );
	// returns what the material score will be after move "m"
        // is made. board holds the position before the move is 
        // made. Score is from the perspective of the current
        // side to move.

	static BOOL check_en_prise( const Board &board, const ColorType attackingSide,
        const Piece p, const Square sq);
	// return "True" if a piece p is unsafe on square "sq".
	
	static int en_prise;
	// set by calls to positional_score or evalu8; count of
	// apparently en prise pieces
		
	static int trapped;
	// set by calls to positional_score or evalu8; true if
	// any piece appears trapped.
		
        static BOOL in_endgame( const Board &board );

        static int king_safety(const Board &, const ColorType,
          King_Pawn_Entry *hash_entry);

        static int king_safety(const Board &);

        static BOOL repetition_draw( const Board &board );

        static BOOL material_draw( const Board &board );
        
        // checks for legal draws plus certain other theoretically
        // draw positions
        static BOOL is_draw(const Board &board);
        
        // Check only for those positions which are not legal draws
        // but are theoretically drawn
        static BOOL is_theoretical_draw(const Board &board);
        
        // Check only for legal draws
        static BOOL is_legal_draw(const Board &board);
protected:
        static int positional_score( const Board &, const ColorType );

};

#endif
