// Copyright 1987-1999 by Jon Dart.  All Rights Reserved.
#if _MSC_VER >= 1200
#pragma optimize("g",off)
#endif

#include "stdafx.h"
#include "bearing.h"
#include "search.h"
#include "scoring.h"
#include "constant.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
#include "pinfo.h"
#include "attacke.h"
#include "log.h"
#ifdef _WINDOWS
   #include "clock.h"
   #include "display.h"
   #include "resource.h"
   #include "mainfrm.h"
#endif
#include "options.h"
#include "globals.h"
#include "notation.h"
#include "pool.h"   
#include "debug.h"
#include "config.h"
#ifdef _WINDOWS
static const UINT SEARCH_TIMER = 2;
#endif
#ifdef _TRACE
   #include <iostream.h>
#endif
#include <limits.h>
#include <process.h>
#include <stdio.h>
#include <wincon.h>

static const int Illegal = -Constants::BIG - 1000;
static int Time_Check_Interval = 1000;

// flags for types of extension
static const int RECAPTURE = 1;
static const int CHECK = 2;
static const int PAWN_PUSH = 4;
static const int FORCED = 8;

#ifdef _WINDOWS
int num_searches = 0;
static Search *searches[MAX_SEARCHES];
#endif

Move pv[Constants::MaxPly][Constants::MaxPly];
int pv_length[Constants::MaxPly];

#define UPDATE_PV(move,ply) \
{ \
   pv[ply][ply] = move; \
   for (int j=ply+1;j<ply+1+pv_length[ply+1];j++) \
     pv[ply][j] = pv[ply+1][j]; \
   pv_length[ply] = pv_length[ply+1]+1; \
/*   cout << "line: ";  \
   for (int j2 = 0; j2 < pv_length[ply]; j2++) \
     cout << MoveImage(pv[ply][j2]) << " "; \
   cout << endl; */ \
}

#ifdef _WINDOWS
void FAR PASCAL TimerProc( HWND hWnd, UINT /*msg*/, UINT wParam,
                           DWORD /*lParam*/)
{
   // timer callback function
   time_t current_time = GetTickCount()/10;
   Search *searcher = searches[wParam-SEARCH_TIMER];
   if (searcher->was_terminated())
      return;
   const Search::Search_Type src_type = searcher->get_search_type();
   if (src_type != Fixed_Ply &&
       current_time - searcher->get_start_time() > searcher->get_time_limit())
   {
      searcher->set_time_up(TRUE);
   }
   the_clock->update( );
   if (!searcher->is_background_search())
   {
      Display::show_search_counts(hWnd,
                                  searcher->get_ply(),searcher->get_numnodes());
      CMainFrame *frame = (CMainFrame*)AfxGetMainWnd();
      frame->update(searcher->get_numnodes());
   }
}
#endif

Search::Statistics::Statistics()
{
   clear();
}

void Search::Statistics::clear()
{
   state = Normal;
   value = 0;
   elapsed_time = 0;
   *best_line_image = '\0';
   for (int i = 0; i < Constants::MaxPly; i++)
     best_line[i] = NullMove;
   num_moves = num_nodes = num_qnodes = depth = 0;
}

Search::Search()
:ply0moves(NULL),ply0scores(NULL),
command_function(NULL),post_function(NULL),terminate_function(NULL),
age(0),hash_init_done(FALSE)
{
#ifdef _WINDOWS
   InitializeCriticalSection(&searchCritSection);
#else
   HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
   DWORD dw;
   pipe = !GetConsoleMode(inh, &dw);
   if (!pipe)
   {
      SetConsoleMode(inh, dw & ~(ENABLE_LINE_INPUT |
                                 ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
   }
#endif   
   ply0moves = new Move[Move_Generator::MaxMoves];
   ply0scores = new int[Move_Generator::MaxMoves];
   for (int i = 0; i < Constants::MaxPly; i++)
      moves_generated[i] = new Move[Move_Generator::MaxMoves];
}

Search::~Search()
{
   cleanup(TRUE);
#ifdef _WINDOWS
   DeleteCriticalSection(&searchCritSection);
#endif
}

void Search::terminate_now(BOOL m)
{
   time_up = terminated = TRUE;
#ifdef _WINDOWS
   makeMove = m;
#endif
}

static BOOL is_move(char *buf)
{
   if (isalpha(*buf) && isdigit(buf[1]) &&
       isalpha(buf[2]) && isdigit(buf[3]))
      return TRUE;
   else if (strcmp(buf,"o-o")==0 || strcmp(buf,"O-O")==0 ||
            strcmp(buf,"o-o-o")==0 || strcmp(buf,"O-O-O") ==0)
      return TRUE;
   else
      return FALSE;
}

#ifndef _WINDOWS
BOOL Search::check_input()
{
   DWORD nchar;
   DWORD dw;
   if (pipe)
   {
      // Use the GnuChess trick of calling PeekNamedPipe to check
      // for user input.  We need to do this because Winboard does
      // not send ^C to interrupt pondering. 
      if (PeekNamedPipe(GetStdHandle(STD_INPUT_HANDLE), NULL, 0, 
                        NULL, &nchar, NULL))
      {
//         if (nchar) cout << "(debug) got console input nchar=" << nchar << endl;
         int nextc;
         BOOL terminate;
         for (unsigned i = 0; i < nchar; i++)
         {
            nextc=getc(stdin);
//            cout << "(debug) got: " << (char)nextc << endl;
            if (nextc == '\n')
            {
               cmd_buf[cmd_buf_count]  = '\0';
               cmd_buf_count = 0;
               // call a function (supplied by the calling code) that
               // sets terminate = TRUE if we should terminate.
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return TRUE;
               }
               break;
            }
            else
               cmd_buf[cmd_buf_count++] = nextc;
         }
         // Hack alert.  Under Windows 95 we sometimes get a move
         // with no trailing carriage return.  Doesn't happen under NT.
         if (cmd_buf_count > 0)
         {
            cmd_buf[cmd_buf_count] = '\0';
            if (is_move(cmd_buf))
            {
//               cout << "(debug) got move w/o <CR>: " << cmd_buf << endl;
               cmd_buf_count = 0;
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return TRUE;
               }
            }
         }
      }
   }
   else
   {
      // We are reading from the console, not connected to Winboard.
      // Not something we usually do, except for testing.
      int retVal = WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), 0);
      if (retVal == WAIT_OBJECT_0)
      {
         INPUT_RECORD recBuf[128];
         ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE),recBuf,
                          128,&dw);
         for (unsigned i = 0; i < dw; i++)
         {
            if (!recBuf[i].Event.KeyEvent.bKeyDown) continue;
            char nextc=recBuf[i].Event.KeyEvent.uChar.AsciiChar;
            if (nextc == (char)13 /* carriage return */)
            {
               putchar('\n');
               cmd_buf[cmd_buf_count]  = '\0';
               cmd_buf_count = 0;
               BOOL terminate;
               // call a function (supplied by the calling code) that
               // sets terminate = TRUE if we should terminate.
               command_function(cmd_buf,terminate);
               if (terminate)
               {
                  return TRUE;
               }
               break;
            }
            else
            {
               putchar(nextc);
               cmd_buf[cmd_buf_count++] = nextc;
            }
         }
      }
   }
   return FALSE;
}
#endif

BOOL Search::check_time(int ply, int rank)
{
   if (terminated)
      return TRUE;
#ifndef _WINDOWS
   if (/*command_function && */check_input())
   {
      terminate_now();
      return TRUE;
   }
#endif
   if (type_of_search != Fixed_Ply)
   {
      int current_time = (int)(GetTickCount()/10);
      if (current_time - (int)get_start_time() > (int)get_time_limit())
      {
         if (talkLevel == Trace)
            terminate_now(TRUE);
         return TRUE;
      }
      if (type_of_search == Fixed_Time)
      {
         return FALSE;
      }
      // Allow terminating early if this is a recapture, appeared
      // better than everything else at initial sort time, and
      // still looks good.
      if (!time_up && easy && iteration_depth >= 5)
      {
         // require that we have consumed at least 1/3 of the time
         // limit:
         current_time = GetTickCount()/10;
         if (time_target >= 100 &&
             (current_time - get_start_time())*3 > time_target)
         {
            if (pv[0][0] == easy_move && 
                PieceMoved(pv[0][0]) != King)
            {
               time_up = TRUE;
               terminate_now(TRUE);
               return TRUE;
            }
         }
      }
   }
   if ( time_up)
   {
      terminate_now(TRUE);
   }
   return time_up;
}

// To reduce the chance of a hash collision, we do some elementary
// testing on moves returned by the hash table, to help weed out
// any that are invalid in the current board position ..
BOOL hash_move_valid( const Board &board, const Move &move)
{
   if (IsNull(move))
      return TRUE;
   if (StartSquare(move) == DestSquare(move))
      return FALSE;
   Piece p = board[StartSquare(move)];
   if (TypeOfPiece(p) != EmptyPiece && PieceColor(p) == board.Side())
   {
      p = board[DestSquare(move)];
      if (IsEmptyPiece(p) || PieceColor(p) != board.Side())
      {
         if (TypeOfMove(move) == EnPassant)
         {
            return board.EnPassantSq(board.OppositeSide()) != InvalidSquare;
         }
         else if (TypeOfMove(move) == KCastle)
         {
            return (board.CastleStatus(board.Side()) ==
                    Board::CanCastleKSide) ||
            (board.CastleStatus(board.Side()) ==
             Board::CanCastleEitherSide);
         }
         else if (TypeOfMove(move) == QCastle)
         {
            return (board.CastleStatus(board.Side()) ==
                    Board::CanCastleQSide) ||
            (board.CastleStatus(board.Side()) ==
             Board::CanCastleEitherSide);
         }
         else
            return TRUE;
      }
      else
         return FALSE;
   }
   else
      return FALSE;
}

void Search::show_status(const Board &board, const char ch)
{
   if (terminated)
      return;
   int ply = stats->depth;
   static char move_text[20];
   Notation::Image(board, pv[0][0]/*move*/, move_text);
   char val[20];
   sprintf(val,"%6.2f",stats->display_value/64.0);
   if (talkLevel == Whisper || talkLevel == Trace)
   {
      if (stats->elapsed_time >= 300 && time_target > 500)
      {
         if (talkLevel == Whisper)
            cout << "tellics whisper ply " << ply;
         else
            cout << "(debug) ply " << ply;
         if (ply < 10) cout << ".  ";
         else cout << ". ";
         cout << move_text << " score: " << val << " " <<
         "time: " << stats->elapsed_time/100 << " secs. nodes: " <<
         stats->num_nodes << endl;
      }
/**
       if ((stats->state == Search::Normal || stats->state == Search::Check) &&
         (stats->value > Constants::BIG-100))
       {
          char text[30];
          if (stats->value != Constants::BIG-1)
             wsprintf(text,"whisper mate in %d!",(Constants::BIG-stats->value-1)/2);
       }
**/
   }
   else if (talkLevel == Debug)
   {
      cout << "ply " << ply;
      if (ply < 10) cout << ".  ";
      else cout << ". ";
      cout << move_text << '\t' << stats->elapsed_time/100 << 
      " seconds.\tscore: " << val  << "\t" <<
      stats->num_nodes << " nodes." << endl;
   }
   if (!is_background_search())
   {
#ifdef _WINDOWS
      CMainFrame *frame = (CMainFrame*)AfxGetMainWnd();
      frame->update(ply,stats->num_nodes,(float)(stats->value/64.0),
                    stats->best_line_image);
#else
      if (post_function) post_function(*stats);
#endif
   }
}

void Search::set_background(BOOL b)
{
   bkgrnd_search = b;
   if (!b)
   {
      // changing from background to foreground search,
      // update display
//      update_stats();
      show_status(initial_board,0);
      // We are in the foreground now, run at normal priority
      SetThreadPriority(GetCurrentThread(),THREAD_PRIORITY_NORMAL);
   }
}

BOOL Search::is_draw( const Board &board, int &rep_count, int ply)
{
   // First check for draw by repetition
   int entries = board.move_count-2;
   hash_t to_match = board.HashCode();
   rep_count = 0;
   for (hash_t *rep_list=board.rep_list_head-3;
       entries>0;
       rep_list-=2,entries-=2)
   {
      if (*rep_list == to_match)
      {
         rep_count++;
         if (rep_count >= 2 ||
             (rep_count == 1 && board.rep_list_head-rep_list <= ply))
         return TRUE;

      }
   }
   // check the 50 move rule
   if (board.move_count >= 100)
      return TRUE;
   // check for insufficient material and other drawing situations
   return Scoring::material_draw(board) || Scoring::is_theoretical_draw(board);
}

int Search::evalu8(const Board & board, int alpha, int beta, int ply)
{
   // Evaluate a board position. 
   const ColorType side = board.Side();
   int rep_count;
   if (is_draw(board,rep_count,ply))
      return draw_score(board);
   int score = Scoring::material_score(board);
   int pos_score = Scoring::positional_score(board,-Constants::BIG,
                                             Constants::BIG);
   score += pos_score;
   return score;
}

int Search::draw_score(const Board &board)
{
   int score = 0;

   // A draw is worse if there is a lot of material on the board
   if (root_total_mat > 6000)
   {
      score -= 25;
   }
   else if (root_total_mat > 4800)
   {
      score -= 25*(root_total_mat-4800)/1200;
   }
   if (board.Side() != side_to_move)
   {
      score = -score;
   }
   int mat_diff = Scoring::material_score(board);
   // Penalize the side to move for allowing a draw if it is ahead in
   // material
   const Material &my_mat = board.getMaterial(board.Side());
   if (Util::Abs(mat_diff) >= 128 &&
       (!my_mat.no_pawns() || my_mat.value() > 2320))
   {
      if (mat_diff > 0)
         score -= (mat_diff-128)/8;
      else
         score -= (mat_diff+128)/8;
   }
#ifdef _TRACE
   cout << "draw score = " << score << endl;
#endif    
   return score;
}

#pragma optimize("g",off)
// calculate a "stand pat" score for the q-search
int Search:: pat_score(const Board &board,int alpha,int beta,int ply)
{
   int rep_count;
   if (is_draw(board,rep_count,ply))
      return draw_score(board);
   int score = Scoring::material_score(board);
   int pos_score;
   pos_score = Scoring::positional_score(board,alpha-score,beta-score);
   score += pos_score;
   return score;
}

int Search::
ply0_move_search(Board & board, int alpha, int beta,
                 int depth,
                 BOOL &terminate)
{
   // implements alpha/beta search for the top most ply.  We use
   // the PVS (principal variation search) variant of alpha/beta.

   time_check_counter++;
   num_nodes++;

   BOOL in_pv = TRUE;
   BOOL in_check = FALSE;
   BOOL stop = FALSE;
   extensions[0] = 0;
   checks[0] = FALSE;
   int move_count = 0;
   int move_index;
   int num_try = 0; // # of legal moves tried
   int i,j;
   current_score = alpha;

/**   if (is_draw(board,rep_count,0) && !Scoring::is_theoretical_draw(board))
   {
#ifdef _TRACE
      cout << "draw!" << endl;
#endif
      current_score = draw_score(board);
      return current_score;
   }**/
   int initial_alpha = alpha;

   Position_Info *hash_entry = NULL;      

   int score = alpha;

   // At this point we need to know if we are in check or not.
   in_check = checks[0] = board.CheckStatus() == Board::InCheck;
   Board_State save_state(board);

   int try_score;
   //
   // sort the moves 
   // pv first, then killer moves
   //
   Move *moves = moves_generated[0];

   Move killer1 = Move_Ordering::get_killer(0,0);
   Move killer2 = Move_Ordering::get_killer(0,1);
   j = 0;
   for (i = 0; i < ply0move_count; i++)
   {
      moves[j++] = ply0moves[i];

      if (MovesEqual(ply0moves[i],ply0first))
      {
         Move tmp = moves[0];
         moves[0] = ply0first;
         moves[i] = tmp;
      }
      else if (MovesEqual(ply0moves[i],killer1))
      {
         int spot = 0;
         if (ply0first != NullMove)
            ++spot;
         Move tmp = moves[spot];
         moves[spot] = ply0first;
         moves[i] = tmp;
      }
      else if (MovesEqual(ply0moves[i],killer2))
      {
         int spot = 0;
         if (ply0first != NullMove)
            ++spot;
         if (killer1 != NullMove)
            ++spot;
         Move tmp = moves[spot];
         moves[spot] = ply0first;
         moves[i] = tmp;
      }
   }

   //
   // Search the next ply
   //
   int newbeta = beta;
   move_index = 0;
   while (move_index < ply0move_count)
   {
      Move rmove = moves[move_index++];
      if (IsMate(rmove)) continue; // move leads to mate, skip it
#ifdef _TRACE
      cout << "window = [" << alpha << ',' << newbeta << ']' <<
      " beta = " << beta << endl;
      cout << "trying " << 0. << ". " << MoveImage(rmove) << endl;
#endif
      last_move[0] = rmove;
      board.DoMove(rmove);
      int extend = 0;
      if (pawn_push_extensions && 
          (PieceMoved(rmove) == Pawn) &&
          (Rank(DestSquare(rmove),board.OppositeSide()) == 7) &&
          (depth > 2*DEPTH_INCREMENT || !board.any_attacks(DestSquare(rmove),board.Side())))
      {
         extend += PAWN_PUSH_EXTENSION;
         extensions[0] |= PAWN_PUSH;
      }
/**      if (check_extensions && in_check)
      {
         extensions[0] |= CHECK;
         extend += CHECK_EXTENSION;
      }
      if (extend >DEPTH_INCREMENT) extend = DEPTH_INCREMENT;**/
      if (depth+extend-DEPTH_INCREMENT >= 0)
      {
         try_score = -move_search(board, -newbeta, -alpha,
                                  1, depth+extend-DEPTH_INCREMENT, in_pv,
                                  terminate);
      }
      else
      {
         try_score = -quiesce(board, -newbeta, -alpha,
                              1, 0);
      }
      if (terminate)
      {
         board.UndoMove(rmove,save_state);
         return alpha;
      }
      // We weeded out illegal moves so should never get one here
      ASSERT(try_score != Illegal);
#ifdef _TRACE
      cout << "0. " << MoveImage(rmove) << ' ';
      cout << try_score;
      if (in_pv) cout << " (pv)";
      cout << endl;
#endif
      num_try++;
      if (!in_pv &&
          ((try_score > alpha) && (try_score < beta)))
      {
         // We failed to get a cutoff and must re-search
#ifdef _TRACE
         cout << "window = [" << alpha << "," << beta
         << "]" << endl;
         cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
         if (depth+extend -DEPTH_INCREMENT >= 0)
            try_score=-move_search(board,
                                   -beta,-alpha,1,depth+extend-DEPTH_INCREMENT,in_pv,terminate);
         else
            try_score=-quiesce(board,
                               -beta,-alpha,1,0);
#ifdef _TRACE
         cout << "0. " << MoveImage(rmove) << ' ';
         cout << try_score;
         cout << endl;
#endif
      }

      board.UndoMove(rmove,save_state);
      if (terminate)
      {
         return alpha;
      }
      ply0scores[move_index-1] = try_score;
      // If this move leads to being mated, remove it from the
      // ply 0 move list - there is no use searching it again
      if (try_score <= -Constants::BIG+(int)iteration_depth)
      {
#ifdef _TRACE
         cout << "removing move " << MoveImage(rmove) << endl;
#endif
         for (int i = 0; i < ply0move_count; i++)
         {
            if (MovesEqual(ply0moves[i],rmove))
            {
               SetMate(ply0moves[i]);
               break;
            }
         }

      }
      if (try_score > alpha)
      {
         current_score = score = try_score;
         if (try_score >= beta)
         {
            best[0] = rmove;
#ifdef _TRACE
			cout << "ply 0 beta cutoff" << endl;
#endif
            if (pv)
            {
               pv[0][0] = rmove;
               pv_length[0] = 1;
            }
            return try_score;
         }
         int pv_change = !MovesEqual(best[0],rmove);
         best[0] = rmove;
         if (pv_change)
         {
            UPDATE_PV(rmove,0);
            update_stats(alpha,beta);
            show_status(board,'>');
         }
         alpha = try_score;
      }
      else if (in_pv && try_score < initial_alpha)
      {
         // we have failed low and must reset the window
         best[0] = rmove;
#ifdef _TRACE
         cout << "ply 0 alpha cutoff" << endl;
#endif
         current_score = try_score;
         return try_score;
      }
      in_pv = FALSE;
      newbeta = alpha + 1; // zero-width window
   }

   if (num_try == 0 && move_count == 0)
   {
      // no moves were tried
      if (in_check)
      {
         if (ply0move_count == 0)        // mate
         {
            score = -(Constants::BIG);
            state = Search::Checkmate;
         }
         else
         {
            // We generated some moves, and some were legal,
            // but we skipped them all.
            score = evalu8(board,alpha,beta,0);
         }
      }
      else    // stalemate
      {
         state = Search::Stalemate;
#ifdef _TRACE
         cout << "stalemate!" << endl;
#endif
         score = draw_score(board);
      }
   }
   else if (num_try > 0)
   {
      if (score >= beta)
      {
         // save move that caused cutoff as a "killer" move:
         Move_Ordering::set_killer(board,last_move[0], 0, depth);
         Move_Ordering::update_history(board,last_move[0],depth);
      }
      else
      {
         // If there's only one legal move, don't search any more:           
         if (num_try == 1 && move_index == move_count)
            terminate = TRUE;
      }
   }
   if (alpha != initial_alpha && score < beta)
   {
      if (num_try == 0)
         pv_length[0] = 0;
      else
      {
         UPDATE_PV(best[0],0);
      }
   }
   ASSERT(score >= -Constants::BIG && score <= Constants::BIG);
   return score;
}

int Search::calc_extensions(const Board &board,int ply,int depth, BOOL in_check, BOOL
                            forced) 
{
   int extend = 0;
   // see if we should apply any extensions at this node.
   extensions[ply] = 0;
   if (forced && forced_extensions && checks[ply])
   {
      extend += FORCED_EXTENSION;
      extensions[ply] |= FORCED;
   }
   else if (check_extensions && in_check && 
            !(ply >= 4 && checks[ply-4] &&
            DestSquare(last_move[ply-4]) == DestSquare(last_move[ply])))
   {
       extensions[ply] |= CHECK;
       extend += CHECK_EXTENSION;
   }
   if (pawn_push_extensions && !IsNull(last_move[ply-1]) && 
       (PieceMoved(last_move[ply]) == Pawn) &&
       (Rank(DestSquare(last_move[ply]),board.OppositeSide()) == 7) &&
       (depth > 2*DEPTH_INCREMENT || !board.any_attacks(DestSquare(last_move[ply]),board.Side())))
   {
      extend += PAWN_PUSH_EXTENSION;
      extensions[ply] |= PAWN_PUSH;
   }
   if (recapture_extensions && ply >= 2 &&
       !(extensions[ply-1] & RECAPTURE) &&
       Capture(last_move[ply]) != EmptyPiece &&
       Capture(last_move[ply-1]) != EmptyPiece &&
       Capture(last_move[ply-1]) != Pawn &&
       DestSquare(last_move[ply-1]) == DestSquare(last_move[ply]))
   {
//      int capVal = PieceValue(Capture(last_move[ply-1]));
//      int capVal2 = PieceValue(Capture(last_move[ply]));
//      if (Util::Abs(capVal - capVal2) < 32)
      {
         extend += RECAPTURE_EXTENSION;
         extensions[ply] |= RECAPTURE;
      }
   }
   return extend;
}

int Search::
quiesce(Board & board, int alpha, int beta,
        const int ply, int depth)
{
   // recursive function, implements quiescence search.
   //
   ASSERT(ply < Constants::MaxPly);
   num_nodes++; time_check_counter++; num_qnodes++;
   pv_length[ply] = 0;
#ifdef CHECKS_IN_QSEARCH
   BOOL in_check;
#endif
   int score = alpha; 
   if (ply >= Constants::MaxPly-1)
   {
      return evalu8(board,alpha,beta,ply);
   }
   int stand_pat_score;
#ifdef CHECKS_IN_QSEARCH
   // At this point we need to know if we are in check or not
   in_check = checks[ply] = (board.CheckStatus() ==
                             Board::InCheck);
#endif

   // Establish a default score.  This score is returned if no
   // captures are generated, or if no captures generate a better
   // score (since we generally can choose whether or not to capture).
   // Normally we don't allow "standing pat" if in check, but we make
   // an exception if we were not in check somewhere earlier in the
   // quiescence search.
   //
#ifdef CHECKS_IN_QSEARCH
   int fwd_prune_ok = 0;
   if (!in_check ||
       (depth <=-2 && !checks[ply-2]) ||
       (depth <=-4 && !checks[ply-4]) ||
       (depth <=-6 && !checks[ply-6]))
   {
      ++fwd_prune_ok;
#else
   {
#endif
#ifdef CHECKS_IN_QSEARCH
      if (checks[ply-1])
         stand_pat_score = pat_score(board,-Constants::BIG,
                                     Constants::BIG,ply);
      else
#endif        
         stand_pat_score = pat_score(board,alpha,beta,ply);

#ifdef _TRACE
      indent(ply);
      cout << "cap score = " << stand_pat_score << endl;
#endif
      if (stand_pat_score > alpha)
      {
         alpha = stand_pat_score;
         if (stand_pat_score >= beta)
         {
#ifdef _TRACE
            indent(ply); cout << "** CUTOFF **" << endl;; 
#endif             
            return stand_pat_score;
         }
      }
   }
#ifdef CHECKS_IN_QSEARCH
   else // standing pat is not an option
      alpha = -Constants::BIG;

#endif
   score = alpha;
   int move_count = 0;
   Move *moves = moves_generated[ply];
   Move_Generator mg(board, ply, NullMove);
   // generate all the capture moves
   move_count = mg.Generate_Captures(moves);
#ifdef _TRACE
   indent(ply); cout << move_count << " capture moves generated" << endl;
#endif
   num_moves += move_count;
   // order captures (and prune losing captures)
   if (move_count > 0)
   {
      move_count = Move_Ordering::order_captures(board, moves, move_count, alpha,
                                                 last_move[ply-1]);
   }
#ifdef _TRACE
   indent(ply); cout << move_count << " capture moves after pruning" << endl;
#endif

   Board_State save_state(board);

   int initial_alpha = alpha;
#ifdef _TRACE
   indent(ply); cout << "alpha = " << alpha << endl;
#endif
   int move_index = 0;
   int num_try = 0; // # of legal moves tried
   int try_score;
   BOOL endgame = Scoring::in_endgame(board);
   best[ply] = NullMove;
   while (move_index < move_count)
   {
      Move emove = moves[move_index++];
      if (Capture(emove) == King)
         return -Illegal;
      // forward prune moves that don't gain enough
#ifdef FORWARD_PRUNE
   #ifdef CHECKS_IN_QSEARCH
      if (fwd_prune_ok && 
   #else
      if (
   #endif
          !endgame)
      {
         if (Capture(emove) != EmptyPiece)
         {
            int est = Scoring::material_score(board,emove);
            if (FORWARD_PRUNE_BIAS + est >= alpha)
            {
               est = attack_estimate(board,emove);
               if ((3*FORWARD_PRUNE_BIAS)/4 + est < alpha)
                  continue;
            }
            else
               continue;
         }
         else //if (TypeOfMove(emove) == Promotion)
         {
            int est = (int)PieceValue(PromoteTo(emove)) - 64;
            if (est + FORWARD_PRUNE_BIAS <= alpha)
            {
               continue;
            }
         }
      }
#endif
#ifdef _TRACE
      indent(ply);
      cout << "trying " << ply << ". " << MoveImage(emove) << endl;
#endif
      last_move[ply] = emove;
      board.DoMoveQ(emove);
      try_score = -quiesce(board, -beta, -alpha, ply+1, depth-1);
      board.UndoMove(emove,save_state);
      if (try_score != Illegal)
      {
#ifdef _TRACE
         indent(ply);
         cout << ply << ". " << MoveImage(emove) << ' ';
         cout << try_score;
         cout << endl;
#endif
         num_try++;
         if (try_score > alpha)
         {
            score = try_score;
            if (try_score >= beta)
               break;
            best[ply] = emove;
            alpha = try_score;
            if (score >= Constants::BIG-1-ply)
               break; // mating move found
         }
      }
   }
   if (score >= beta)
   {
#ifdef _TRACE
      indent(ply);
      cout << "**CUTOFF**" << endl;
#endif 
      return score;
   }
#ifdef CHECKS_IN_QSEARCH
   #ifdef _TRACE
   indent(ply);
   cout << "num_try = " << num_try << endl;
   #endif 
   if (num_try == 0 && move_index == move_count
       && in_check)
   {
      // no moves were tried.  Make sure we did not place
      // ourselves in check to get to this position.
      if (board.any_attacks(board.KingPos(
                                         board.OppositeSide()),board.Side()))
         score = -Illegal;
      else if (move_count == 0)
      {
         // no legal moves, mate
         score = -(Constants::BIG - ply);
      }
      else
      {
         // We are in check, but all check-avoidance
         // moves have been forward pruned.  So we
         // need to calculate a score.
         if (score == -Constants::BIG)
            score = pat_score(board,-Constants::BIG,Constants::BIG,ply);
      }
      if (score > alpha && score < beta)
         alpha = score;
   }
#endif
   if (alpha != initial_alpha && score < beta)
   {
      if (best[ply] == NullMove)
         pv_length[ply] = 0;
      else
      {
         UPDATE_PV(best[ply],ply);  
         Move_Ordering::update_history(board,best[ply],depth);
      }
   }
   return score;
}

Move Search::find_best_move(
                           CWnd* parentWin,
                           const Board &board, 
                           Search_Type srcType,
                           int32 timelimit,
                           int32 xtratime,
                           int plylimit,
                           const BOOL background,
                           Statistics &stat_buf,
                           Talk_Level t)
{
   talkLevel = t;
   time_target = timelimit;
   xtra_time = xtratime;
   ply_limit = plylimit;
   type_of_search = srcType;
   stats = &stat_buf;
   stats->clear();

   bkgrnd_search = background;
   Board board_copy(board);
   initial_board = board;

   init(board_copy);
#ifdef _WINDOWS
   my_hwnd = parentWin->m_hWnd;
   EnterCriticalSection(&searchCritSection);
   ASSERT(num_searches < MAX_SEARCHES);
   int search_number = num_searches;
   searches[search_number] = this;
   ++num_searches;

   // set timer every 1000 ms. (1 second). 
   idTimer = parentWin->SetTimer(num_searches+SEARCH_TIMER-1, 
                                 1000, (TIMERPROC)TimerProc);
   if (!idTimer)
   {
      parentWin->MessageBox("Too many clocks or timers!","",
                            MB_ICONEXCLAMATION | MB_OK);
   }
   LeaveCriticalSection(&searchCritSection);
#endif    
   // Do the actual search
   iterate(board_copy);

   static const BOOL end_of_game[] = { FALSE, FALSE, FALSE, TRUE, TRUE,
      TRUE, TRUE, TRUE, TRUE};
   if (!end_of_game[(int)stats->state] && global_options->can_resign())
   {
      if (stats->depth >= 2)
      {
         // Use the display value - i.e. do not use score if
         // we are failing low
         if (stats->display_value <= 8-(int)Constants::BIG)
         {
            // TBD we can sometimes get -10000 as a score!
            // this is bogus, we should not resign.
            if (stats->display_value != -(int)Constants::BIG)
               stats->state = Resigns;
         }
         else if (stats->display_value > 100-Constants::BIG &&
                  (stats->display_value <= (int)Constants::Contempt &&
                   Scoring::positional_score(board,-Constants::BIG,Constants::BIG) < 64))
            stats->state = Resigns;
      }
   }

   if (talkLevel == Debug)
   {
      if (stats->elapsed_time > 0)
         cout << (100.0*(float)stats->num_nodes)/((float)stats->elapsed_time) << " nodes/second." << endl;
      cout << (num_nodes-num_qnodes) << " regular nodes, " <<
      num_qnodes << " quiescence nodes." << endl;
      cout << hash_searches << " searches of hash table, " <<
      hash_hits << " successful (" << 
      (int)((100.0*(float)hash_hits)/((float)hash_searches)) <<
      " percent)." << endl;
      cout << hash_inserts << " hash entries inserted, " <<
      hash_replaces << " entries replaced, " <<
      hash_inserts_failed << " inserts failed." << endl;
      fflush(stdout);
   }
#ifdef _WINDOWS
   EnterCriticalSection(&searchCritSection);
   --num_searches;
   LeaveCriticalSection(&searchCritSection);
   KillTimer(my_hwnd,idTimer);
#endif
   return pv[0][0];
}

void Search::update_stats(int alpha, int beta)
{
   time_t end_time = GetTickCount()/10;
   stats->elapsed_time = end_time - start_time;
   stats->num_nodes = num_nodes;
   stats->num_qnodes = num_qnodes;
   stats->num_moves = num_moves;
   // Note: the search routine counts depth internally starting
   // at 0, but we report it starting at 1 (as most other programs do).
   stats->depth = iteration_depth+1;
   stats->value = current_score;
   // Leave the previous value intact if we have failed low.
   if (!(stats->value == alpha && iteration_depth > 0))
   {
      stats->display_value = stats->value;
   }
   stats->state = state;
   if (terminated)
   {
      // keep the previous "best line" intact.
      return;
   }
   *(stats->best_line_image) = '\0';
   stats->best_line[0] = NullMove;

   Board board_copy(initial_board);
 
   if (IsNull(pv[0][0]))
   {
#ifdef _TRACE
       cout << "warning: pv is null\n";
#endif       
       return;
   }
   else if (pv_length[0] == 0)
   {
       return;
   }
   int i = 0;
   while (i < pv_length[0] && !IsNull(pv[0][i]))
   {
      stats->best_line[i] = pv[0][i];
      char move_image[20];
      *move_image = '\0';
      Notation::Image(board_copy,stats->best_line[i],move_image);
      strcat(stats->best_line_image,move_image);
      strcat(stats->best_line_image," ");
      board_copy.DoMove(stats->best_line[i]);
      ++i;
      int rep_count;
      if (is_draw(board_copy,rep_count,0))
      {
         break;
      }
      if (pv_length[0] < 2 && use_hash_table)
      {
         // get the next move from the hash table, if possible
         // (for pondering)
         Position_Info *hash_entry = searchHash(board_copy,rep_count);
         if (hash_entry)
         {
            Move hashMove = hash_entry->best_move(board_copy);
            if (!IsNull(hashMove))
            {
               stats->best_line[i] = hashMove;
               *move_image = '\0';
               Notation::Image(board_copy,hashMove,move_image);
               strcat(stats->best_line_image,move_image);
               ++i;
            }
         }
         break;
      }
   }

   stats->best_line[i] = NullMove;
#ifdef _TRACE
   cout << "best line: " << stats->best_line_image << endl;
#endif
}

void Search::iterate(const Board &board)
{

   BOOL terminate = FALSE;
   BOOL research;
   if (ply_limit > 0)
      ply_limit = Util::Min(ply_limit-1,Constants::MaxPly-1);

   // Generate the ply 0 moves here:
   Move_Generator mg(board,0,NullMove);
   ply0move_count = mg.Generate_Moves(ply0moves);
   num_moves += ply0move_count;
   // Searching may alter the board, so operate on a copy:
   Board board_copy(board);
   // Order the moves.  The ordering process removes any illegal moves
   // at ply 0, so fix the move count if necessary.
   ply0move_count = 
   Move_Ordering::order_ply0_moves(board_copy, ply0moves, ply0move_count, 
                                   NullMove,
                                   easy);
   // See if there is an "easy" move (a recapture that appears to
   // gain more material than any other alternative).  If so we
   // may be able to terminate the search early.
   easy &= (ply0move_count > 0 &&
            board[DestSquare(ply0moves[0])] != EmptyPiece &&
            game_moves->num_moves() > 0 &&
            DestSquare(game_moves->move(game_moves->num_moves()-1))
            == DestSquare(ply0moves[0]));
   easy_move = ply0moves[0];
   // Positions are stored in the hashtable with an "age" to identify
   // which search they came from.  "Newer" positions can replace 
   // "older" ones.  Update the age here since we are starting a
   // new search.
   age = (age + 1) % 3;
   // Incrementally search the board to greater depths, up to the
   // ply limit
   for (iteration_depth = 0;
       iteration_depth <= ply_limit && !terminate && !time_up; 
       iteration_depth++)
   {
      int value;
      research = FALSE;
      int lo_window, hi_window;
      if (iteration_depth == 0)
         current_score = value = Scoring::evalu8(board);
      else
         value = stats->value;
      lo_window = Util::Max(-Constants::BIG,value - Constants::Initial_Search_Window / 2);
      hi_window = Util::Min(value + Constants::Initial_Search_Window / 2,
                            Constants::BIG);
      BOOL failhigh = FALSE;
      BOOL faillow = FALSE;
      while (1)
      {
         int root_beta = hi_window;
#ifdef _TRACE
         cout << "root_beta = " << root_beta << endl;
#endif
         if (!failhigh && !faillow)
            ply0first = best[0];
         ply0_move_search(board_copy, lo_window, hi_window, 
                          DEPTH_INCREMENT*iteration_depth, terminate);
         update_stats(lo_window,hi_window);
         if (iteration_depth >= 2 && (ply0move_count == 1 ||
            Scoring::is_legal_draw(board)))
         {
            // Draw, or there is only one legal move, terminate.
            terminate = TRUE;
         }
         if (stats->elapsed_time > 100)
         {
            Time_Check_Interval = (int)(500L*stats->num_nodes)/(36L*stats->elapsed_time);
         }
         if (!terminate)
         {
            terminate = check_time(0,0);
            if (terminate_function && terminate_function(*stats))
            {
               terminate = TRUE;
            }
         }
         if (terminate)
         {
            show_status(board,0);
            break;
         }
         else
         {
            show_status(board,0);
#ifdef _TRACE
            cout << "stats->value = " << stats->value << "  root_beta = "
            << root_beta << endl;
#endif
            ASSERT(stats->value != Constants::BIG);
            if (stats->value >= root_beta)
            {
#ifdef _TRACE
               cout << "ply 0 high cutoff, re-searching ..." << endl;
#endif
               // high cutoff, must re-search
#ifdef _TRACE
               cout << "last_move[0] = " << MoveImage(last_move[0]) << endl;
#endif
               ply0first = last_move[0]; // so we research this move
               hi_window = Constants::BIG;
               lo_window = Util::Min(stats->value,Constants::BIG-Constants::MaxPly);
               failhigh = TRUE;
               faillow = FALSE;
            }
            else if (stats->value <= lo_window && !failhigh)
            {
#ifdef _TRACE
               cout << "ply 0 low cutoff, re-searching ..." << endl;
               cout << "value = " << stats->value << endl;
#endif
               hi_window = lo_window;
               lo_window = -Constants::BIG;
               if (iteration_depth > 3 && type_of_search == Time_Limit &&
                   stats->elapsed_time > time_target/2 &&
                   stats->value < best_scores[iteration_depth-1]-48 &&
                   time_added == 0)
               {
                  // search more time because our score is dropping
                  time_added = xtra_time;
                  if (talkLevel == Trace)
                     cout << "(debug) adding time" << endl;
               }
               failhigh = FALSE;
               faillow = TRUE;
            }
            else
            {
               break; // out of loop
            }
         }
         best_moves[iteration_depth] = best[0];
         best_scores[iteration_depth] = stats->value;
         research = TRUE;
      }
#ifdef _TRACE
      cout << iteration_depth+1 << " ply search result: " << MoveImage(pv[0][0]) << 
      "   value = " << stats->value << endl;
#endif
      if (stats->value <= 1 + (int)(iteration_depth*2) - Constants::BIG)
      {
         // We're either checkmated or we certainly will be, so
         // quit searching.
#ifdef _TRACE
         cout << "terminating, low score" << endl;
#endif            
         break;
      }
      else if (stats->value >= (Constants::BIG - (int)(iteration_depth*2)
                                  - 1))
      {
#ifdef _TRACE
         cout << "terminating, mate score" << endl;
#endif            
         break;
      }
   }
   if (Scoring::is_legal_draw(board))
   {
       stats->state = Search::Draw;
       stats->value = draw_score(board);
   }
}


int Search::get_ply() const
{
   // return current depth of search.
   return iteration_depth;
}

unsigned long Search::get_numnodes() const
{
   // return # of nodes visited
   return num_nodes;
}    

void Search::init(const Board &board)
{
   num_nodes = num_qnodes = num_moves = hash_hits = 
               hash_searches = hash_inserts = hash_replaces = hash_inserts_failed = 0L;
   time_check_counter = 0;
   time_up = FALSE;
   terminated = FALSE;
   time_added = 0;
   iteration_depth = 0;
   current_score = 0;
   state = Normal;
   side_to_move = board.Side();
   root_total_mat = board.getMaterial(White).value() + board.getMaterial(Black).value();
   start_time = GetTickCount()/10;
   for (int i = 0; i< Constants::MaxPly;i++)
   {
      pv[i][0] = best[i] = last_move[i] = best_moves[i] = NullMove;
      extensions[i] = checks[i] = 0;
      pv_length[i] = 0;
   }
   ply0first = NullMove;
   Scoring::init(board);
   Move_Ordering::clear_killer();
   Move_Ordering::clear_history();

   SearchOps src_opts;
   global_options->get_search_ops(src_opts);
   null_depth = src_opts.null_depth*DEPTH_INCREMENT;
   check_extensions = src_opts.check_extensions;
   pawn_push_extensions = src_opts.pawn_push_extensions;
   forced_extensions = src_opts.forced_extensions;
   recapture_extensions = src_opts.recapture_extensions;

   // Allocate the hash table.  Size is in kilobytes.  Guesstimate
   // a conversion to the number of hash table entries. 
   use_hash_table = src_opts.hash_table_size > 0;
   int hash_entries;
   hash_entries = (1000*src_opts.hash_table_size)/sizeof(Position_Info);
#ifdef _WINDOWS
   if (!hash_init_done && use_hash_table)
   {
      if (src_opts.auto_size_hash_table)
      {
         long size = calc_hash_size();
         initHash(size);
         hash_init_done = TRUE;
      }
      else if (hash_entries >0)
      {
         initHash(hash_entries);
         hash_init_done = TRUE;
      }
   }
   makeMove = TRUE;
#else
   // no auto size option for console program
   if (!hash_init_done && use_hash_table && src_opts.hash_table_size >0)
   {
      initHash(hash_entries);
      hash_init_done = TRUE;
   }
#endif    
}

void Search::reinit()
{
   // After the search options change, we may need to reallocate
   // the hash table.  Free it here so that next time through init()
   // the reallocation will be done.
   freeHash();
   hash_init_done = FALSE;
//   Position_Info::freeAll(FALSE);
}

long Search::calc_hash_size()
{
   // Size the hash table according to how much memory we have.
   static int sizes[12]={ 997, 1997, 3001, 4003, 4999, 6007, 6997, 8003,
      10007, 12007, 15013, 18103};

   DWORD memSize = GlobalCompact(0);
   int i;
   for (i = 0; i < 11; i++)
      if (sizeof(Position_Info)*((long)sizes[i]) > memSize/16)
         break;
   return 32L*sizes[i];
}

void Search::cleanup(BOOL clearHash)
{
   if (hash_init_done && clearHash)
   {
      freeHash();
   }
   delete [] ply0moves;
   delete [] ply0scores;
   for (int i = 0; i< Constants::MaxPly;i++)
   {
      if (moves_generated[i] != NULL)
         delete [] moves_generated[i];
   }
}

struct SearchParams
{
   SearchParams(Search *searcher,
                CWnd *, const Board &,
                Search::Search_Type srcType,
                int32 time_limit, int32 xtra_time,
                int ply_limit,
                const BOOL background,
                Search::Statistics &,
                Search::Talk_Level t);
   Search *searcher;
   CWnd* parentWin;
   const Board &ABoard; 
   Search::Search_Type type_of_search;
   int32 time_limit, xtra_time;
   int ply_limit;
   Search::Statistics &stats;
   BOOL background;
   Search::Talk_Level talkLevel;
};

SearchParams::SearchParams(
                          Search *srch,
                          CWnd* pWin,
                          const Board &b,
                          Search::Search_Type src_type,
                          int32 timelimit,
                          int32 xtratime,
                          int plylimit,
                          const BOOL bkgrnd,
                          Search::Statistics &s,
                          Search::Talk_Level tl)
: searcher(srch),parentWin(pWin),ABoard(b),type_of_search(src_type),
time_limit(timelimit),xtra_time(xtratime),
ply_limit(plylimit),background(bkgrnd),
stats(s), talkLevel(tl)
{
}

// Thread procedure
static unsigned int _stdcall threadProc(LPVOID pParam)
{
   SearchParams *params = (SearchParams*)pParam;
   params->searcher->find_best_move(params->parentWin,
                                    params->ABoard,
                                    params->type_of_search,
                                    params->time_limit,
                                    params->xtra_time,
                                    params->ply_limit,
                                    params->background,
                                    params->stats,
                                    Search::Silent);
   // signal the parent that our search is done
#ifdef _WINDOWS
   ((CArasanView*)params->parentWin)->OnSearchComplete(
        params->searcher->make_the_move());
#endif      
   delete params;
   return 0;
}

#ifdef _WINDOWS
HANDLE Search::start_search(
                           CWnd* parentWin,
                           const Board &board, 
                           Search_Type srcType,
                           int32 time_limit,
                           int32 xtratime,
                           int ply_limit,
                           const BOOL background,
                           Statistics &stats,
                           Search::Talk_Level t)
{
   // pack the parameters into a single structure
   SearchParams *pParam = new SearchParams(
                                          this,
                                          parentWin,board,srcType,time_limit,xtratime,ply_limit,
                                          background,stats, t);
   // start the thread.  If this is a background search, use
   // IDLE priority to keep the GUI responsive.
   unsigned int threadId;
   HANDLE hThread = (HANDLE)_beginthreadex(NULL,0 /*stack*/,
                                           &threadProc,(void*)pParam, (DWORD)0,
                                           &threadId);
   #ifdef _WINDOWS
   if (hThread)
      SetThreadPriority(hThread,background ? THREAD_PRIORITY_BELOW_NORMAL : THREAD_PRIORITY_NORMAL);
   #endif
   return hThread;
}
#endif



