// Copyright 1994, 1997 by Jon Dart.  All Rights Reserved.

#ifndef _SQUARE_H
#define _SQUARE_H

#include "types.h"
#include "color.h"

// Square used to be a class but now it is typedef'd to an int,
// for efficiency reasons. This header defines several convenience
// macros and functions for dealing with squares.

typedef unsigned int Square;

enum {InvalidSquare = 127};

extern const int Files[64];
extern const int Ranks[64];
extern const int Edge[64];
extern const int Colors[64];
extern const int Center[64];
extern const int Corners[64];

#define Square( file, rank, side ) \
  ((side == Black) ? ((rank-1)*8 + file - 1) : (55 - (rank-1)*8 + file))
    
// returns "rank" from the perspective of "side".  1 = 1st
// rank for side, 8 = last rank.	    
#define Rank(square, side ) ((side == Black) ? Ranks[square] : 9-Ranks[square])

// returns file.  Queen rook file = 1, King rook file = 8.
#define File(square) Files[square]
	    
#define OnBoard(square) ((square & ~0x3f) == 0)

#define OnEdge(square) Edge[square]

#define InCenter(square) Center[square]

#define InCorner(square) Corners[square]
    
#define IsInvalid(square) (square == InvalidSquare)
    
#define SquareColor(square) ((ColorType)Colors[square])
    
// Parse a square value in algebraic notation (e.g. "g7") and
// return a corresponding Square type.
extern Square SquareValue(char *p);
    
extern char FileImage(Square sq);

extern char RankImage(Square sq);

#endif

